from golgi.config import PathConfigOption
from golgi.config.options import DictConfigOption

metadata = dict(parents=['series'])


def reset_config():
    return {
        'etvdb': dict(
            path=PathConfigOption('/usr/bin/etvdb',
                                  help='etvdb executable path'),
            series_name_map=DictConfigOption(),
        )
    }


__all__ = ['reset_config', 'metadata']



import re

from series.get.model.show import Show

from amino import Maybe
from series import canonicalize
from series.get.tvdb import Tvdb

from series.db_facade import DbFacade, exclusive, commit


class ShowsFacade(Tvdb, DbFacade):

    @property
    def main_type(self):
        return Show

    @exclusive
    def name_exists(self, name):
        return self.filter_by(canonical_name=canonicalize(name)).count() > 0

    def id_param(self, showid):
        return (dict(etvdb_id=showid) if self.use_etvdb else
                dict(rage_id=showid))

    @exclusive
    def add(self, name, show):
        data = dict(
            name=show.name,
            canonical_name=canonicalize(name)
        )
        data.update(**self.id_param(show.showid))
        if show.latest_episode is not None:
            data.update(latest_episode=show.latest_episode.number,
                        latest_season=show.latest_episode.season)
        self._db.add(Show(**data))

    @commit
    def update(self, id_, data):
        query = self.filter_by(id=id_)
        if query.count() > 0:
            show = query.first()
            for key, value in data.items():
                setattr(show, key, value)
            return show
        else:
            self.log.debug(
                'Tried to update nonexistent show with id {}'.format(id_)
            )

    @commit
    def delete_by_sid(self, showid):
        query = self.filter_by(**self.id_param(showid))
        if query.count() > 0:
            show = query.first()
            self._db.delete(show)
        else:
            self.log.debug(
                'Tried to delete nonexistent show with id {}'.format(showid)
            )

    @commit
    def delete(self, showid):
        query = self.filter_by(id=showid)
        if query.count() > 0:
            show = query.first()
            self._db.delete(show)
        else:
            self.log.debug(
                'Tried to delete nonexistent show with id {}'.format(showid)
            )

    @property  # type: ignore
    @exclusive
    def all(self):
        return self.filter_by().all()

    def filter_by_regex(self, regex):
        r = re.compile(regex)
        return filter(lambda s: r.search(s.canonical_name), self.all)

    @exclusive
    def find_by_name(self, name):
        return Maybe(self.filter_by(canonical_name=name).first())

__all__ = ['ShowsFacade']



from typing import Callable, Any

from flask import request

from golgi.config import configurable
from tek.tools import is_int

from amino import Empty, Just, __, Map, List, Either, F

from series.rest_api import RestApi as RestApiBase, route_decorator
from series.get.tvdb import Tvdb
from series.get.model.show import Show
from series.get import ReleaseMonitor


def error(msg, status=404):
    return dict(error=msg), status


@configurable(get=['rest_api_host', 'rest_api_port'])
class RestApi(RestApiBase, Tvdb):

    routes, route = route_decorator()

    def __init__(self, releases, shows, *a, **kw):
        RestApiBase.__init__(self, 'get')
        self._releases = releases
        self._shows = shows
        self._getd = Empty()

    def set_getd(self, daemon):
        self._getd = Just(daemon)

    def _no_release(self, _id):
        return error('No release with id {} found'.format(_id))

    def _release_map(self, _id, f: Callable[[ReleaseMonitor], Any]):
        release = self._releases.find_by_id(_id)
        if release:
            return f(release)
        else:
            return self._no_release(_id)

    def _release_foreach(self, _id, f: Callable[[ReleaseMonitor], Any]):
        def go(r):
            f(r)
            return 'ok'
        return self._release_map(_id, go)

    def _no_show(self, _id):
        return error('No show with id {} found'.format(_id))

    def _show_map(self, _id, f: Callable[[Show], Any]):
        show = self._shows.find_by_id(_id)
        if show:
            return f(show)
        else:
            return self._no_show(_id)

    @property
    def _no_getd(self):
        return error('SeriesGetD not running or connected to rest api', 500)

    @property
    def _json(self):
        return Map(request.get_json())

    def _either(self, result: Either, status=404):
        return result.map(lambda a: "ok").right_or_map(F(error, status=status))

    @route('/release/explain_old')
    @route('/release/explain_old/<int:num>')
    def explain_old(self, num=5):
        return [r.explain_downloadable for r in self._releases[-num:]]

    @route('/release/<int:_id>/link', methods=['PUT'])
    def add_link(self, _id):
        url = self._json['url']
        if self._releases.add_link_by_id(_id, url):
            return 'Done'
        else:
            return self._no_release(_id)

    @route('/download/pending')
    def pending_downloads(self):
        releases = self._releases.pending_downloads
        dls = [(r.id, r.release.search_string) for r in releases]
        return dls

    @route('/release/<series>/<season>/<episode>/downloaded', methods=['PUT'])
    def mark_downloaded_metadata(self, series, season, episode):
        monitor = self._releases.find_by_metadata(series, season, episode)
        return self._mark_downloaded(monitor)

    @route('/release/<int:_id>/downloaded', methods=['PUT'])
    def mark_downloaded_id(self, _id):
        monitor = self._releases.find_by_id(_id)
        return self._mark_downloaded(monitor)

    @route('/release')
    def release_list(self):
        regex = self._json.get_or_else('regex', '')
        matches = self._releases.filter_episode_repr(regex)
        return [[str(monitor.id), str(monitor.release)] for monitor in matches]

    @route('/release_id')
    def id_by_metadata(self):
        data = self._json
        results = self._releases.find_by_metadata(data['series'],
                                                  data['season'],
                                                  data['episode'])
        return results.id if results else -1

    @route('/release/<series>/<season>/<episode>', methods=['POST'])
    def create_release(self, series, season, episode):
        if self._releases.find_by_metadata(series, season, episode):
            return 'Release already exists'
        else:
            airdate = (self._shows.find_by_name(series)
                       .map(lambda s: self.tvdb.airdate(s, season, episode))
                       .get_or_else(None)
                       )
            self._releases.create(series, season, episode, airdate=airdate)
            return 'ok'

    @route('/release/<series>/<season>/<episode>', methods=['DELETE'])
    def delete_release(self, series, season, episode):
        if self._releases.find_by_metadata(series, season, episode):
            self._releases.delete(series, season, episode)
            return 'ok'
        else:
            return 'Release doesn\'t exist'

    @route('/release/<int:_id>', methods=['PUT'])
    def update_release(self, _id):
        if self._releases.find_by_id(_id):
            return self._releases.update_by_id(_id, **self._json).info
        else:
            return 'Release doesn\'t exist'

    @route('/season', methods=['POST'])
    def add_season(self):
        data = self._json
        if 'name' in data and 'season' in data and is_int(data['season']):
            name, season_name = data['name'], int(data['season'])
            show = self.tvdb.show(name)
            if show:
                season = self.tvdb.season(show.name, season_name)
                if season:
                    count = len(season)
                    self._releases.add_season(name, season_name, count)
                    return 'Done.'
                else:
                    return 'No such season: {} s{}'.format(name, season_name)
            else:
                return 'No such show: {}'.format(name)
        else:
            return 'Specify name and season in the parameters!'

    @route('/show', methods=['DELETE'])
    def delete_show(self):
        data = self._json
        if 'name' in data:
            name = data['name']
            if name.isdigit():
                self._shows.delete(int(name))
                return 'Done'
            else:
                show = self.tvdb.show(name)
                if show:
                    self._shows.delete_by_sid(show.showid)
                    return 'Done'
                else:
                    return 'No show with name "{}"'.format(name)
        else:
            return 'Specify show name in the parameters!'

    @route('/show', methods=['POST'])
    def add_show(self):
        data = self._json
        if 'name' in data:
            name = data['name']
            show = self.tvdb.show(name)
            if show:
                self._shows.add(name, show)
                return 'Done'
            else:
                return 'Adding show {} failed'.format(name)
        else:
            return 'Specify show name in the parameters!'

    @property
    def _target_shows(self):
        regex = self._json.get_or_else('regex', '')
        return self._shows.filter_by_regex(regex)

    @route('/show')
    def list_shows(self):
        return [[str(s.id), str(s.name)] for s in self._target_shows]

    def current_release(self, show):
        return self._releases.one(show.canonical_name, show.current_season,
                                  show.current_episode)

    def next_info(self, show):
        status = 0
        if show.has_next_episode:
            nepi = 'next episode: {}'.format(show.next_episode_date.date())
        else:
            nepi = 'no next episode'
            relinfo = None
        release = self.current_release(show)
        if release:
            if release.downloaded:
                status = 1
                air = release.release.airdate
                date = ' ({})'.format(air.date()) if air else ''
                enum = 'x'.join(map(str, release.enum))
                relinfo = 'release for {} downloaded{}'.format(enum, date)
            elif release.torrent:
                if release.torrent_valid():
                    valid = ''
                    status = 2
                else:
                    valid = 'in'
                    status = 3
                relinfo = 'release with {}valid torrent'.format(valid)
            else:
                relinfo = 'release without torrent'
        else:
            relinfo = 'no release yet'
        return [show.name, nepi, relinfo, status]

    @route('/show/info')
    def show_info(self):
        return list(map(self.next_info, self._target_shows))

    @route('/show/next')
    def next_shows(self):
        shows = [s for s in self._target_shows if s.has_next_episode]
        return list(map(self.next_info, shows))

    @route('/show/ready')
    def ready_shows(self):
        def has_release(show: Show):
            r = self.current_release(show)
            return (r is not None and
                    r.release.airdate == show.next_episode_date)
        shows = [s for s in self._target_shows if s.has_next_episode]
        valid = filter(has_release, shows)
        return list(map(self.next_info, valid))

    @route('/show/done')
    def done_shows(self):
        def done(show: Show):
            r = self.current_release(show)
            return r is not None and r.downloaded
        valid = filter(done, self._target_shows)
        return list(map(self.next_info, valid))

    def _mark_downloaded(self, monitor):
        if monitor:
            msg = 'Marking release "{}" as downloaded.'
            self.log.info(msg.format(monitor.release))
            self._releases.update_by_id(monitor.id, downloaded=True)
            self._releases.commit()
        return monitor is not None

    def _activate(self, _id):
        self._getd.foreach(__.activate_release(_id))

    @route('/release/<int:_id>/reset_torrent', methods=['PUT'])
    def reset_torrent(self, _id):
        def go(r):
            data = self._json
            self._releases.reset_torrent(_id)
            if 'link' in data and data['link']:
                self._releases.add_link_by_id(_id, data['link'])
            self._activate(_id)
        return self._release_foreach(_id, go)

    @route('/release/<int:_id>', methods=['GET'])
    def release_show(self, _id):
        release = self._releases.find_by_id(_id)
        return release.info if release is not None else self._no_release(_id)

    @route('/release/<int:_id>/activate', methods=['PUT'])
    def activate_release(self, _id):
        go = lambda r: self._activate(_id)
        return self._release_foreach(_id, go)

    @route('/release/<int:_id>/explain', methods=['GET'])
    def explain_release(self, _id):
        def go(release):
            services = self._json.get('services') / __.split(',') | ['all']
            s = List.wrap(services)
            r = List(release)
            return self._getd / __.explain_release(r, s) | self._no_getd
        return self._release_map(_id, go)

    @route('/show/<int:_id>/explain', methods=['GET'])
    def explain_show(self, _id):
        def go(show):
            services = self._json.get('services') / __.split(',') | ['all']
            s = List.wrap(services)
            r = List(show)
            return self._getd / __.explain_show(r, s) | self._no_getd
        return self._show_map(_id, go)

    @route('/release/<int:_id>/airdate', methods=['PUT'])
    def set_airdate(self, _id):
        return self._either(
            self._json.get('date')
            .to_either('no date given') //
            F(self._releases.set_airdate, _id),
            500
        )

__all__ = ['RestApi']



import os
import socket

import requests

from series.logging import Logging
from golgi.config import configurable
from tek.errors import NotEnoughDiskSpace
from tek.tools import sizeof_fmt

from tek_utils.sharehoster import DownloaderFactory
from tek_utils.sharehoster.errors import ShareHosterError

from amino import List

from series.get.handler import ReleaseHandler, R, L
from series.condition import DynOr, HandlerCondition
from series.get.model.link import Link, TorrentLink
from series.get.model.release import ReleaseMonitor


class ValidLink(HandlerCondition):

    def __init__(self, link: Link) -> None:
        self.link = link

    @property
    def cond(self):
        if isinstance(self.link, TorrentLink):
            return self._cond_torrent
        else:
            return self._cond_http

    @property
    def _cond_torrent(self):
        return (~(L('url', None) | L('dead') | L('download_url', None)) &
                L('cached'))

    @property
    def _cond_http(self):
        return L('valid')

    def ev(self, item):
        return self.cond.ev(self.link)

    def describe(self, item, target):
        d = self.cond.describe(self.link, target)
        m = self.link.url
        return '{} -> {}'.format(m, d)


class ValidLinks(DynOr):

    @property
    def _sub_type(self):
        return ValidLink

    def _dyn_subs(self, item: ReleaseMonitor):
        return List.wrap(item.links.all())

    def describe(self, item, target):
        d = super().describe(item, target)
        return 'links => {}'.format(d)


@configurable(get=['download_dir', 'min_size'])
class Downloader(ReleaseHandler):

    def __init__(self, releases, *a, **kw):
        super().__init__(releases, 5, 'download manager')
        self._min_failed = 0
        self._downloader_fact = DownloaderFactory()

    def _fail(self, monitor, link):
        f = monitor.failed_downloads + 1
        self._update(monitor, failed_downloads=f)
        self._releases.fail_link(link.id)

    def _handle(self, monitor):
        self.log.info('Downloading {!s}.'.format(monitor.release))
        self._update(monitor, downloading=True)
        link = monitor.link
        self._download_dir.mkdir(parents=True, exist_ok=True)
        try:
            dl = self._downloader_fact(link.download_url,
                                       download_dir=str(self._download_dir))
            dl.retrieve()
        except NotEnoughDiskSpace as e:
            self.log.error('Downloader: {}'.format(e))
        except (requests.RequestException, ShareHosterError,
                ConnectionError, socket.timeout) as e:
            self._fail(monitor, link)
            self.log.error('Downloader: {}'.format(e))
        else:
            self._check_download(monitor, link, dl)
        finally:
            self._update(monitor, downloading=False)

    def _check_download(self, monitor, link, dl):
        ''' Verify the downloaded file's size by means of two criteria:
        The content-length header sent from the hoster must match
        exactly.
        The config parameter min_size defines what is considered to be a
        faulty file, i.e. the link checker didn't detect the file to be
        invalid, but it is small enough to just be the hoster's error
        page.
        '''
        size = dl.outfile_size
        if size < self._min_size:
            self._fail(monitor, link)
            msg = 'File smaller than required ({}): {!s}'
            self.log.error(msg.format(sizeof_fmt(size), monitor.release))
        elif size != dl.file_size and dl.file_size != -1:
            self._fail(monitor, link)
            msg = 'Filesize mismatch in download of {!s}:'
            self.log.error(msg.format(monitor.release))
            msg = 'Got {}, expected {}.'
            self.log.error(msg.format(size, dl.file_size))
            try:
                dl.file_path.unlink()
            except IOError:
                msg = 'Couldn\'t delete broken download of {}.'
                self.log.warn(msg.format(monitor.release))
        else:
            self._releases.update_by_id(monitor.id,
                                        download_path=str(dl.file_path))
            self._releases.mark_episode_downloaded(monitor)

    # @property
    # def _candidates(self):
    #     candidates = self._releases.downloadable
    #     counts = [r.failed_downloads for r in candidates]
    #     self._min_failed = min(counts or [0])
    #     return candidates

    @property
    def _conditions(self):
        return (~(R('downloaded') | R('nuked') | R('downloading')) &
                ValidLinks())

    def _choose(self, candidates):
        counts = [r.failed_downloads for r in candidates]
        min_failed = min(counts or [0])
        best = candidates.filter(lambda a: a.failed_downloads == min_failed)
        return best.head

    # def _qualify(self, release):
    #     return release.failed_downloads == self._min_failed

__all__ = ['Downloader']



"""add WatchEvent.stopped_at column

Revision ID: 1f69c57af9a
Revises: 26134e26286
Create Date: 2014-09-16 02:32:38.448495

"""

# revision identifiers, used by Alembic.
revision = '1f69c57af9a'
down_revision = '26134e26286'

from alembic import op
import sqlalchemy as sa


def upgrade():
    ### commands auto generated by Alembic - please adjust! ###
    op.add_column('watch_event', sa.Column('stopped_at', sa.Float(), nullable=True))
    ### end Alembic commands ###


def downgrade():
    ### commands auto generated by Alembic - please adjust! ###
    op.drop_column('watch_event', 'stopped_at')
    ### end Alembic commands ###
